/* This file is Copyright 2002 Level Control Systems.  See the included LICENSE.txt file for details. */  

#include <new.h>
#include <stdio.h>
#include "regex/StringMatcher.h"
#include "util/String.h"

#include <string.h>

namespace muscle {

StringMatcher::StringMatcher() : _regExpValid(false), _rangeMin(MUSCLE_NO_LIMIT), _rangeMax(MUSCLE_NO_LIMIT)
{
   // empty
} 

StringMatcher :: StringMatcher(const char * str, bool simple) : _regExpValid(false)
{
   SetPattern(str, simple);
}

StringMatcher :: ~StringMatcher()
{
   if (_regExpValid) regfree(&_regExp);
}

status_t StringMatcher::SetPattern(const char * str, bool isSimple) 
{
   String pattern;

   _rangeMin = _rangeMax = MUSCLE_NO_LIMIT;  // default to regular matching mode
   if (isSimple)
   {
      // Special case for strings of form e.g. "<15-23>", which is interpreted to
      // match integers in the range 15-23, inclusive.  Yeah, it's an ungraceful
      // hack, but also quite useful since regex won't do that in general.
      if (str[0] == '<')
      {
         const char * dash = strchr(str+1, '-');
         if (dash)
         {
            const char * nulByte = strchr(dash+1, '\0');
            if (*(nulByte-1) == '>')
            {
               _rangeMin = (str[1]  == '-') ? 0 : (uint32)atoi(str+1);
               _rangeMax = (dash[1] == '>') ? MUSCLE_NO_LIMIT : ((dash[1] == '-') ? 0 : (uint32)atoi(dash+1));
            }
         }
      }

      if (_rangeMin == MUSCLE_NO_LIMIT)
      {
         if ((str[0] == '\\')&&(str[1] == '<')) str++;  // special case escape of initial < for "\<15-23>"

         pattern = "^\\(";

         bool escapeMode = false;
         for (const char * ptr = str; *ptr != '\0'; ptr++)
         {
            if (escapeMode)
            { 
               escapeMode = false;
               switch(*ptr)
               {
                  case ',': case '|': case '(': case ')': case '?': 
                     pattern += *ptr;
                  break;
      
                  default:
                     pattern += '\\';
                     pattern += *ptr;
                  break;
               }
            }
            else
            {
               switch(*ptr) 
               {
                  case ',': case '|': 
                     pattern += "\\|";
                  break;

                  case '.': case '(': case ')': 
                     pattern += '\\';
                     pattern += *ptr;
                  break;

                  case '*': 
                     pattern += ".*";
                  break;

                  case '?': 
                     pattern += '.';
                  break;

                  case '\\': 
                     escapeMode = true;
                  break;

                  default:  
                     pattern += *ptr;
                  break;
               }
            }
         }
         pattern += "\\)$";
         //printf("OUTPUT: pattern became '%s'.\n", pattern.Cstr());
      }
   }

   // Free the old regular expression, if any
   if (_regExpValid)    
   {
     regfree(&_regExp);
     _regExpValid = false;
   }

   // And compile the new one
   if (_rangeMin == MUSCLE_NO_LIMIT)
   {
      _regExpValid = (regcomp(&_regExp, (pattern.Length() > 0) ? pattern.Cstr() : str, 0) == 0);
      return _regExpValid ? B_NO_ERROR : B_ERROR;
   }
   else return B_NO_ERROR;  // for range queries, we don't need a valid regex
}


bool StringMatcher::Match(const char * const str) const
{
   if (_rangeMin == MUSCLE_NO_LIMIT)
   {
      if (_regExpValid == false) return false;
      int regExpStat = regexec(&_regExp, str, 0, NULL, 0);
      return (regExpStat != REG_NOMATCH);
   }
   else
   {
      uint32 val = (uint32) atoi(str);
      return ((val >= _rangeMin)&&(val <= _rangeMax));
   }
} 


bool IsRegexToken(char c)
{
   switch(c)
   {
     case '[': case ']': case '*': case '?': case '\\': case ',': case '|': case '(': case ')':
        return true;

     default:
        return false;
   }
}

void EscapeRegexTokens(String & s)
{
   const char * str = s.Cstr();

   String ret;

   if (*str == '<') ret += '\\';  // special case escape for "<15-23>" case
   while(*str)
   {
     if (IsRegexToken(*str)) ret += '\\';
     ret += *str;
     str++;
   }
   s = ret;
}

bool HasRegexTokens(const char * str)
{
   while(*str)
   {
     if (IsRegexToken(*str)) return true;
                        else str++;
   }
   return false;
}

bool MakeRegexCaseInsensitive(String & str)
{
   bool changed = false;
   String ret;
   for (uint32 i=0; i<str.Length(); i++)
   {
     char next = str[i];
     if ((next >= 'A')&&(next <= 'Z'))
     {
        char buf[5];
        sprintf(buf, "[%c%c]", next, next+('a'-'A'));
        ret += buf;
        changed = true;
     }
     else if ((next >= 'a')&&(next <= 'z'))
     {
        char buf[5];
        sprintf(buf, "[%c%c]", next, next+('A'-'a'));
        ret += buf;
        changed = true;
     }
     else ret += next;
   }
   if (changed) str = ret;
   return changed;
}

};   // end namespace muscle
