/* This file is Copyright 2002 Level Control Systems.  See the included LICENSE.txt file for details. */  

#ifndef MuscleMemoryAllocator_h
#define MuscleMemoryAllocator_h

#include "support/MuscleSupport.h"
#include "util/Queue.h"
#include "util/RefCount.h"

namespace muscle {

class MemoryAllocator;

/** Interface class representing an object that can allocate and free blocks of memory. */
class MemoryAllocator : public RefCountable
{
public:
   /** Default constructor; no-op */
   MemoryAllocator();

   /** Virtual destructor, to keep C++ honest */
   virtual ~MemoryAllocator();

   /** Should allocate a block of (howMuch) bytes and return it, or return NULL on failure. */
   virtual void * Malloc(uint32 howMuch) = 0;

   /** Should free a previously-allocated block (memory).  
     * Note that passing in a NULL pointer is not allowed, and may result in undefined behaviour.
     */
   virtual void Free(void * memory) = 0;

   /** Should return true iff Malloc() has ever failed (i.e. returned NULL).
     * @param setFlagTo Value that this method should return the next time it is called. 
     */
   virtual bool HasMallocFailed(bool setFlagTo) const = 0;
};

typedef Ref<MemoryAllocator> MemoryAllocatorRef;

/** This is a very simple, no-frills memory allocator that just calls malloc() and delete() */
class BasicMemoryAllocator : public MemoryAllocator 
{
public:
   /** Default constructor */
   BasicMemoryAllocator();

   /** Destructor */
   virtual ~BasicMemoryAllocator();

   virtual void * Malloc(uint32 howMuch);
   virtual void Free(void * memory);
   virtual bool HasMallocFailed(bool setFlagTo) const;

private:
   mutable bool _hasMallocFailed;
};

/** Convenience class, used for easy subclassing:  holds a slave MemoryAllocator and passes all method
  * calls on through to the slave.
  */
class ProxyMemoryAllocator : public MemoryAllocator
{
public:
   /** Constructor.
     * @param slaveRef Reference to another MemoryAllocator that we will pass all our method calls on through to.
     */
   ProxyMemoryAllocator(MemoryAllocatorRef slaveRef) : _slaveRef(slaveRef) {/* empty */}

   /** Destructor */
   virtual ~ProxyMemoryAllocator() {/* empty */}

   virtual void * Malloc(uint32 howMuch) {return _slaveRef()->Malloc(howMuch);}
   virtual void Free(void * memory) {_slaveRef()->Free(memory);}
   virtual bool HasMallocFailed(bool setFlagTo) const {return _slaveRef()->HasMallocFailed(setFlagTo);}

private:
   MemoryAllocatorRef _slaveRef;
};

/** This MemoryAllocator decorates its slave MemoryAllocator to 
  * enforce a user-defined limit on how much memory may be allocated at any given time. 
  */
class UsageLimitProxyMemoryAllocator : public ProxyMemoryAllocator
{
public:
   /** Constructor.  
     * @param slaveRef Reference to a sub-MemoryAllocator whose methods we will call through to.
     * @param maxBytes The maximum number of bytes that we will allow (slave) to allocate.  Default to no limit.
     *                 This value may be reset later using SetMaxNumBytes().
     * @param currentBytes The current number of bytes that we should consider to be allocated.  Defaults to zero.
     */
   UsageLimitProxyMemoryAllocator(MemoryAllocatorRef slaveRef, uint32 maxBytes = MUSCLE_NO_LIMIT, uint32 currentBytes = 0);

   /** Destructor.  */
   virtual ~UsageLimitProxyMemoryAllocator();

   virtual void * Malloc(uint32 howMuch);
   virtual void Free(void * memory);

   /** Returns the number of bytes that are currently available to be allocated */
   uint32 GetNumAvailableBytes() const {return (_maxBytes > _currentBytes) ? (_maxBytes-_currentBytes) : 0;}

   /** Returns the maximum number of bytes that may be allocated at any given time */
   uint32 GetMaxNumBytes() const {return _maxBytes;}

   /** Returns the number of bytes that are currently allocated */
   uint32 GetNumUsedBytes() const {return _currentBytes;}

   /** Set a new maximum number of allocatable bytes */
   void SetMaxNumBytes(uint32 mb) {_maxBytes = mb;}

private:
   uint32 _maxBytes;
   uint32 _currentBytes;
};

/** Interface class representing an object that can be called by an AutoCleanupProxyMemoryAllocator
  * in case of an out-of-memory emergency.
  */
class OutOfMemoryCallback : public RefCountable
{
public:
   /** Constructor */
   OutOfMemoryCallback();

   /** Virtual destructor, to keep C++ honest */
   virtual ~OutOfMemoryCallback();

   /** Called by a MemoryAllocator when it is having trouble allocating memory.
     * Typically this method would be implemented to free up some unimportant memory, so that
     * the MemoryAllocator will have a better chance of being able to get memory when it tries again.
     */
   virtual void OutOfMemory() = 0;
};

/** Convenience type for referencing an OutOfMemoryCallback */
typedef Ref<OutOfMemoryCallback> OutOfMemoryCallbackRef;

/** A handy little class--instead of having to subclass, it will call the function you specify */
class FunctionOutOfMemoryCallback : public OutOfMemoryCallback
{
public:
   /** Signature of the function type that we know how to call */
   typedef void (*OutOfMemoryCallbackFunc)();   

   /** Constructor.
     * @param f The function to call if we run out of memory 
     */
   FunctionOutOfMemoryCallback(OutOfMemoryCallbackFunc f) : _func(f) {/* empty */}

   /** Calls the function specified in the header */
   virtual void OutOfMemory() {if (_func) _func();}

private:
   OutOfMemoryCallbackFunc _func;
};

/** This MemoryAllocator decorates its slave MemoryAllocator to call a list of
  * OutOfMemoryCallback objects when the slave's memory allocation fails.  These
  * OutOfMemoryCallback calls should try to free up some memory if possible;
  * then this MemoryAllocator will call the slave again and see if the memory
  * allocation can now succeed.
  */
class AutoCleanupProxyMemoryAllocator : public ProxyMemoryAllocator
{
public:
   /** Constructor.  
     * @param slaveRef Reference to a sub-MemoryAllocator whose methods we will call through to.
     */
   AutoCleanupProxyMemoryAllocator(MemoryAllocatorRef slaveRef);

   /** Destructor.  Calls ClearCallbacks(). */
   virtual ~AutoCleanupProxyMemoryAllocator();

   virtual void * Malloc(uint32 howMuch);

   /** Add a new callback object to our list of callbacks to call during an emergency.
     * @param cbRef Reference to the new callback item
     * @returns B_NO_ERROR on success, or B_ERROR on failure (out of memory)
     */
   status_t AddCallback(OutOfMemoryCallbackRef cbRef);

   /** Clears our list of OutOfMemoryCallback items */
   void ClearCallbacks();

private:
   Queue<OutOfMemoryCallbackRef> _callbacks;
};

};  // end namespace muscle

#endif
